"""Test for the criticalPower module"""

from src.cycling_dynamics.critical_power import CriticalPower


def test_convert_user_critical_power() -> None:
    """Test the conversion of user input to a dataframe and dictionary
    This happens on class initiation if the user provides a critical power profile."""
    user_input = """1, 1000
    5, 800
    30, 500
    60, 450
    300, 400
    1200, 350"""
    profile = user_input.split("\n")
    profile = [x.split(",") for x in profile]
    profile = {int(x[0]): int(x[1]) for x in profile}
    cpp = CriticalPower(cp_user=profile)
    df = cpp.cp_defined_df
    cp = cpp.cp_defined_dict
    # df, cp = convert_user_critical_power(profile)
    assert {1: 1000, 5: 800, 30: 500, 60: 450, 300: 400, 1200: 350}.items() <= cp.items()
    assert df["seconds"].max() == 1200
    assert df["power"].max() == 1000
    assert df["seconds"].min() == 1
    assert df["power"].min() == 350


def test_ramp_test_activity() -> None:
    user_input = """1, 1000
    5, 800
    30, 500
    60, 450
    300, 400
    1200, 350"""
    profile = user_input.split("\n")
    profile = [x.split(",") for x in profile]
    profile = {int(x[0]): int(x[1]) for x in profile}
    cpp = CriticalPower(cp_user=profile)
    df, dfwko = cpp.ramp_test_activity()
    assert df["power"].max() == 1000
    assert df["power"].min() == 350
    assert dfwko["power"].max() == 1000


def test_critical_power_read_fit() -> None:
    """Test the reading of a fit file"""
    FIT_FILE = "test_data/vincent_lap_1_24HOP_14012433014_ACTIVITY.fit"
    cpp = CriticalPower(activity=FIT_FILE)
    cpp.calculate_cp()
    correct_values = [
        (1, 680.0),
        (5, 550.2),
        (10, 515.2),
        (20, 446.05),
        (30, 397.7),
        (60, 331.3666666666667),
        (120, 280.34166666666664),
        (300, 261.0833333333333),
        (600, 258.47),
        (1200, 246.69916666666666),
    ]
    for point in correct_values:
        assert cpp.cp_points[point[0]].cp == point[1]


def test_critical_power_cp_intensity() -> None:
    FIT_FILE = "test_data/vincent_lap_1_24HOP_14012433014_ACTIVITY.fit"
    cpp = CriticalPower(activity=FIT_FILE)
    cpp.cp_intensity()


def test_ramp_test_activity():
    user_input = """1, 1000
    5, 800
    30, 500
    60, 450
    300, 400
    1200, 350"""
    profile = user_input.split("\n")
    profile = [x.split(",") for x in profile]
    profile = {int(x[0]): int(x[1]) for x in profile}
    cpp = CriticalPower(cp_user=profile)
    df, df_wko = cpp.ramp_test_activity()
    assert df["power"].max() == 1000
    assert df["power"].min() == 350
    assert df_wko["power"].max() == 1000


def test_make_zwo_from_ramp():
    user_input = """1, 1000
    5, 800
    30, 500
    60, 450
    300, 400
    1200, 350"""
    profile = user_input.split("\n")
    profile = [x.split(",") for x in profile]
    profile = {int(x[0]): int(x[1]) for x in profile}
    cpp = CriticalPower(cp_user=profile)
    df, df_wko = cpp.ramp_test_activity()
    wko1 = cpp.make_zwo_from_ramp(df_wko, filename=None, name="test", ftp=250)
    assert wko1 is not None
    assert wko1.startswith("<?xml version='1.0' encoding='UTF-8'?>")
    assert wko1.endswith("</workout_file>\n")



"""Python module to contain base calculating functions"""

from math import asin, atan, cos, radians, sin, sqrt, tan


def haversine_distance(lat1: float, lon1: float, lat2: float, lon2: float) -> float:
    """
    Calculate great circle distance between two points in a sphere,
    given longitudes and latitudes https://en.wikipedia.org/wiki/Haversine_formula

    We know that the globe is "sort of" spherical, so a path between two points
    isn't exactly a straight line. We need to account for the Earth's curvature
    when calculating distance from point A to B. This effect is negligible for
    small distances but adds up as distance increases. The Haversine method treats
    the earth as a sphere which allows us to "project" the two points A and B
    onto the surface of that sphere and approximate the spherical distance between
    them. Since the Earth is not a perfect sphere, other methods which model the
    Earth's ellipsoidal nature are more accurate but a quick and modifiable
    computation like Haversine can be handy for shorter range distances.

    Args:
        lat1, lon1: latitude and longitude of coordinate 1
        lat2, lon2: latitude and longitude of coordinate 2
    Returns:
        geographical distance between two points in metres
    >>> from collections import namedtuple
    >>> point_2d = namedtuple("point_2d", "lat lon")
    >>> SAN_FRANCISCO = point_2d(37.774856, -122.424227)
    >>> YOSEMITE = point_2d(37.864742, -119.537521)
    >>> f"{haversine_distance(*SAN_FRANCISCO, *YOSEMITE):0,.0f} meters"
    '254,352 meters'
    """
    # CONSTANTS per WGS84 https://en.wikipedia.org/wiki/World_Geodetic_System
    # Distance in metres(m)
    AXIS_A = 6378137.0
    AXIS_B = 6356752.314245
    RADIUS = 6378137
    # Equation parameters
    # Equation https://en.wikipedia.org/wiki/Haversine_formula#Formulation
    flattening = (AXIS_A - AXIS_B) / AXIS_A
    phi_1 = atan((1 - flattening) * tan(radians(lat1)))
    phi_2 = atan((1 - flattening) * tan(radians(lat2)))
    lambda_1 = radians(lon1)
    lambda_2 = radians(lon2)
    # Equation
    sin_sq_phi = sin((phi_2 - phi_1) / 2)
    sin_sq_lambda = sin((lambda_2 - lambda_1) / 2)
    # Square both values
    sin_sq_phi *= sin_sq_phi
    sin_sq_lambda *= sin_sq_lambda
    h_value = sqrt(sin_sq_phi + (cos(phi_1) * cos(phi_2) * sin_sq_lambda))
    return 2 * RADIUS * asin(h_value)


def angle_type(a, b, c):
    """
    Calculate the cosine of the angle using Law of Cosines and determine the angle type
    Is the angle between sides a and b acute, obtuse or 90 degrees (right angle)
     a: The original control point to next point on path
     b: The original control point to point on other path
     c: Second point on path the other path point.
    """
    angle = (a**2 + b**2 - c**2) / (2 * a * b)

    # Determine the angle type
    if angle > 0:  # After control point
        return "after"
    elif angle < 0:  # Before control point
        return "before"
    else:
        return "beside"  # cos_C == 0 implies a right angle



import logging

import pandas as pd
from garmin_fit_sdk import Decoder, Stream

logging.basicConfig(level=logging.INFO)


def load_fit_file(file_path, add_metrics=True, rolling_window=30) -> pd.DataFrame:
    logging.info("Loading FIT file")
    stream = Stream.from_file(file_path)
    decoder = Decoder(stream)
    if not decoder.is_fit():
        logging.error(f"File is not FIT: {file_path}")
        raise f"File is not FIT: {file_path}"
    messages, errors = decoder.read()
    if errors:
        logging.error(f"Errors: {errors}")
        raise f"Errors: {errors}"
    df = pd.DataFrame(messages["record_mesgs"])
    df["position_lat"] = df["position_lat"] / 1e7
    df["position_long"] = df["position_long"] / 1e7
    if "enhanced_speed" in df.columns and "speed" in df.columns:
        logging.info("Using enhanced speed")
        df.drop(columns=["speed"], inplace=True)
    df.rename(columns={"enhanced_speed": "speed"}, inplace=True)
    if "enhanced_altitude" in df.columns and "altitude" in df.columns:
        logging.info("Using enhanced altitude")
        df.drop(columns=["altitude"], inplace=True)
    df.rename(columns={"enhanced_altitude": "altitude"}, inplace=True)
    return df



import logging

import pandas as pd

from src.cycling_dynamics.calc import haversine_distance

logging.basicConfig(level=logging.INFO)


def add_metrics(df: pd.DataFrame, rolling_window: int = 30, ftp: int | None = None) -> pd.DataFrame:
    """Add metrics to the dataframe"""
    logging.info("Adding metrics")
    # Speed
    df[f"speed {rolling_window}sec"] = df["speed"].rolling(window=rolling_window).mean()

    # Efficiency
    df[f"speed per watt {rolling_window}sec"] = (
        df["speed"].rolling(window=rolling_window).mean() / df["power"].rolling(window=rolling_window).mean()
    )
    df[f"speed sqrd per watt {rolling_window}sec"] = (
        df["speed"].rolling(window=rolling_window).mean() ** 2 / df["power"].rolling(window=rolling_window).mean()
    )

    # power
    df["np"] = (df["power"] ** 4).rolling(window=30).mean() ** 0.25
    if ftp is not None:
        df["IF"] = df["np"] / ftp
        df["TSS"] = (df["power"] * df["IF"] * df["seconds"] / ftp / 3600).cumsum()
    return df


def create_grouped_segments(
    tracks: list[pd.DataFrame, ...], start_distance: float, length: float
) -> list[pd.DataFrame]:
    """Compare multiple activities over a segment.
    Tracks: 2 or more activities in the form of a dataframe.
    Track[0] is treated as the control/base track.
    start_point: The "distance" withing track[0] to use as the control point. You can use the exact or the closest will
    be found.
    length: The length of the segment from the control points
    """
    # start_idx = control[control['distance']==start_point]['distance'].sub(start_point).abs().idxmin()
    start_idx = tracks[0]["distance"].sub(start_distance).abs().idxmin()
    logging.info(f"Start index: {start_idx}")
    end_idx = tracks[0]["distance"].sub(start_distance + length).abs().idxmin()
    logging.info(f"End index: {start_idx}")
    start_point = tracks[0].iloc[start_idx][["position_lat", "position_long", "distance"]].values.tolist()
    logging.info(f"Start point lat, lon, distance: {start_point}")
    end_point = tracks[0].iloc[end_idx][["position_lat", "position_long", "distance"]].values.tolist()
    logging.info(f"End point lat, lon, distance: {end_point}")

    # create trimmed track
    control = tracks[0].loc[start_idx:end_idx].copy()
    control.reset_index(drop=True, inplace=True)
    logging.info(f"Distance: {control['distance'].min()}, {control['distance'].max()}")
    control["distance"] = control["distance"] - start_point[2]
    logging.info(f"Distance: {control['distance'].min()}, {control['distance'].max()}")
    control["seconds"] = control["timestamp"].sub(control["timestamp"].min()).dt.total_seconds()
    control["ride"] = 0

    trimmed_tracks = [control]
    for i, track in enumerate(tracks[1:]):
        logging.info(f"Track {i + 1}")
        track["distance_to_start"] = track.apply(
            lambda row: haversine_distance(start_point[0], start_point[1], row["position_lat"], row["position_long"]),
            axis=1,
        )
        match_start_idx = track["distance_to_start"].idxmin()
        logging.info(f"Match start index: {match_start_idx}")
        match_start_point = track.iloc[match_start_idx][["position_lat", "position_long", "distance"]].values.tolist()
        logging.info(f"Match start point lat, lon, distance: {match_start_point}")

        track["distance_to_end"] = track.apply(
            lambda row: haversine_distance(end_point[0], end_point[1], row["position_lat"], row["position_long"]),
            axis=1,
        )
        match_end_idx = track["distance_to_end"].idxmin()
        logging.info(f"Match start index: {match_end_idx}")
        match_end_point = track.iloc[match_end_idx][["position_lat", "position_long", "distance"]].values.tolist()
        logging.info(f"Match end point lat, lon, distance: {match_end_point}")

        trimmed_track = track.loc[match_start_idx:match_end_idx].copy()
        trimmed_track.reset_index(drop=True, inplace=True)
        trimmed_track["distance"] = trimmed_track["distance"] - match_start_point[2]
        trimmed_track["ride"] = i + 1
        trimmed_track["seconds"] = trimmed_track["timestamp"].sub(trimmed_track["timestamp"].min()).dt.total_seconds()
        trimmed_tracks.append(trimmed_track)

    return trimmed_tracks


def normalized_power(df: pd.DataFrame) -> pd.DataFrame:
    """Calculate the normalized power of a ride"""
    df["np"] = (df["power"] ** 4).rolling(window=30).mean() ** 0.25
    return df["np"]


def intensity_factor(df: pd.DataFrame, ftp: int) -> pd.DataFrame:
    """Calculate the intensity factor of a ride"""
    df["IF"] = ((df["power"] ** 4).rolling(window=30).mean() ** 0.25) / ftp
    return df



"""A collection of plots for visualizing cycling dynamics."""

import pandas as pd
import plotly.express as px
import plotly.graph_objects as go


def plot_critical_power_intensity(
    df, width: int = 15, intervals: list[int, ...] = [15, 30, 60, 120, 300, 600, 900, 1200]
) -> px.bar:
    """Plot the critical power intensity
    df: this is the dataframe from get_critical_power_intensity(df, cp=None)
    """
    if "seconds" not in df.columns:
        df["seconds"] = df.index + 1
    number_of_intervals = len(intervals)
    diff_cols_bar = [f"percent_cp_{interval}sec" for interval in intervals]
    diff_cols_bar.reverse()

    every_x_row = df.iloc[::width].copy()
    every_x_row["percent_total_cp"] = every_x_row["percent_total_cp"] * number_of_intervals
    fig = px.bar(
        every_x_row,
        x="seconds",
        y=diff_cols_bar,
        title="Critical Power Intensity",
        labels={"value": "Percent Critical Power", "variable": "% Critical Power"},
        color_discrete_sequence=px.colors.sequential.Plasma,
    )
    fig.add_scatter(
        x=every_x_row["seconds"],
        y=every_x_row["percent_total_cp"],
        mode="lines",
        name="percent_total_cp",
        line=dict(color="red", width=2),
    )
    return fig


def plot_activity_critical_power(df: pd.DataFrame) -> go.Figure:
    """Plot the critical power of an activity
    df: with df being the critical power dataframe from critical_power(df)['df']
    returns: fig
    """
    fig = go.Figure()

    upper_std = df["cp"] + df["std"]
    lower_std = df["cp"] - df["std"]

    fig.add_trace(
        go.Scatter(x=df["seconds"], y=df["cp"], fill=None, mode="lines", line_color="red", name="Critical Power")
    )
    fig.add_trace(
        go.Scatter(x=df["seconds"], y=upper_std, fill="tonexty", mode="lines", line_color="lightblue", name="STD")
    )
    fig.add_trace(
        go.Scatter(x=df["seconds"], y=lower_std, fill="tonexty", mode="lines", line_color="lightblue", name="STD")
    )
    fig.update({"layout": {"title": "Critical Power", "xaxis": {"title": "Time (s)"}, "yaxis": {"title": "Power (W)"}}})

    return fig



"""Critical power related code"""

import logging
import warnings
from dataclasses import asdict, dataclass

import pandas as pd
from cycling_dynamics.load_data import load_fit_file

logging.basicConfig(level=logging.INFO)


@dataclass
class CPPoint:
    """Dataclass to hold critical power data"""

    seconds: int
    idx: int  # index location of the cp in the dataframe
    cp: float
    std: float
    max: float
    min: float
    slope: float
    window: pd.DataFrame  # copy of the window data
    extra_cols: list[str]  # list of extra columns to calculate
    extra_data: dict[str:list]  # dict of data. Column names with list of values.
    intensity: float = 0
    chr: float = 0
    chr_std: float = 0
    chr_max: float = 0
    chr_min: float = 0


def _calculate_ramp_power(df: pd.DataFrame) -> pd.DataFrame:
    """Calculate the ramp power for each second
    df is a dataframe of a 1sec resolution CP curve"""
    df["ramp_power"] = 0.0
    for idx in df.index:
        if idx == 0:
            df.loc[idx, "ramp_power"] = df.loc[idx, "power"]
        else:
            df.loc[idx, "ramp_power"] = df.loc[idx, "power"] * df.loc[idx, "seconds"] - df.loc[: idx - 1][
                "ramp_power"
            ].sum().clip(min=0)
    return df


def _interpolate_curve(df: pd.DataFrame) -> pd.DataFrame:
    """Interpolate the power curve to 1 second intervals"""
    max_time = df["seconds"].max()
    df.set_index("seconds", inplace=True)
    # fill in missing seconds
    new_index = pd.Index(range(1, max_time + 1), name="seconds")
    df = df.reindex(new_index)
    # Interpolate power for missing seconds
    df["power"] = df["power"].interpolate(method="linear")
    df.reset_index(inplace=True)
    return df


class CriticalPower:
    """Critical Power class
    activity: Activity dataframe, file path or None
    cp_user: dict of duration: power, User defined critical power or None. This gets converted to a dataframe and dict
    with 1sec resolution
    max_window: int, the maximum duration to calculate, interpolate critical power, default is 1200 seconds
    """

    def __init__(
        self,
        activity: pd.DataFrame | str | None = None,
        cp_user: dict[int, float] | None = None,
        max_window: int = 1200,
    ):
        self.activity = load_fit_file(activity) if isinstance(activity, str) else activity
        self.max_window = max_window
        self.cp_points: dict[int, CPPoint] = {}
        self.cp_df: pd.DataFrame = pd.DataFrame()
        self.cp_defined_df = None  # this is the user defined CP
        self.cp_defined_dict = None  # this is the user defined CP
        self.ramp_test_df: pd.DataFrame | None = None
        self.ramp_test_wko: pd.DataFrame | None = None
        self.activity_percent_cp: float | None = None
        self.df_rolling: pd.DataFrame | None = None

        if cp_user is not None:
            self._convert_cp_defined(cp_user)

    def _convert_cp_defined(self, cp_user: dict[int, float]):
        """Convert a user defined critical power to a dict and Dataframe interpolated to every second.
        Returns: a df and a dict"""
        logging.info("Convert cp_defined critical power")
        try:
            assert cp_user[1] >= 0
        except AssertionError as err:
            logging.error(f"Critical power must be greater than 0: {err}")
            raise AssertionError("The profile must start with 1 second") from err
        logging.info(f"Last max window of profile: {max(cp_user.keys())}")

        df = pd.DataFrame([(k, v) for k, v in cp_user.items()], columns=["seconds", "power"])
        self.cp_defined_df = _interpolate_curve(df)
        self.cp_defined_dict = self.cp_defined_df.set_index("seconds")["power"].to_dict()

    def calculate_cp(self):
        """Calculate the critical power of an activity"""
        logging.info(f"Calculate critical power up to {self.max_window} seconds")
        df_data = []
        for s in range(1, self.max_window + 1):
            # Power based CP
            self.activity[f"{s}_mean"] = self.activity["power"].rolling(window=s).mean()
            if not self.activity[f"{s}_mean"].isnull().all():
                idx = self.activity[f"{s}_mean"].idxmax()
                window = self.activity.loc[idx - s + 1 : idx].copy().reset_index(drop=True)
                cp_w = window["power"].mean()
                std_w = window["power"].std()
                max_w = window["power"].max()
                min_w = window["power"].min()
                slope_w = window["power"].loc[s // 2 : s].mean() - window["power"].loc[: s // 2].mean()
                try:
                    chr_w = window["heart_rate"].mean()
                    chr_std_w = window["heart_rate"].std()
                    chr_max_w = window["heart_rate"].max()
                    chr_min_w = window["heart_rate"].min()
                except Exception as err:
                    logging.warning("Could not calculate heart rate metrics")
                    logging.warning(err)
                    chr_w = 0
                    chr_std_w = 0
                    chr_max_w = 0
                    chr_min_w = 0

                cols = [
                    col
                    for col in window.columns
                    if col
                    not in ["seconds", "power", "timestamp", "position_lat", "position_long", f"{s}_mean", "heart_rate"]
                    and not isinstance(col, int)
                ]
                extra_data = {}
                for c in cols:
                    try:
                        extra_data[f"{c}_mean"] = window[c].mean()
                        extra_data[f"{c}_std"] = window[c].std()
                        extra_data[f"{c}_max"] = window[c].max()
                        extra_data[f"{c}_min"] = window[c].min()
                    except Exception as err:
                        logging.warning(f"Could not calculate metrics for {c}")
                        logging.warning(err)
                cpp = CPPoint(
                    seconds=s,
                    idx=idx,
                    window=window,
                    cp=cp_w,
                    std=std_w,
                    max=max_w,
                    min=min_w,
                    slope=slope_w,
                    extra_cols=cols,
                    extra_data=extra_data,
                    chr=chr_w,
                    chr_std=chr_std_w,
                    chr_max=chr_max_w,
                    chr_min=chr_min_w,
                )
                self.cp_points[s] = cpp
                df_data_row = {
                    k: v for k, v in asdict(cpp).items() if k in ["seconds", "idx", "cp", "std", "max", "min", "slope"]
                }
                df_data_row.update(extra_data)
                df_data.append(df_data_row)
            self.activity.drop(columns=f"{s}_mean", inplace=True)
        self.cp_df = pd.DataFrame(df_data)

    def get_power_roll_avg_df(self, window: int = 1200) -> pd.DataFrame:
        """Add rolling average power to the dataframe
        df: dataframe with a power column
        """
        max_window = min(window, len(self.activity))
        rolling_power = [
            self.activity["power"].rolling(window=rolling_window, min_periods=1).mean()
            for rolling_window in range(1, max_window + 1)
        ]
        rolling_power_names = [f"power_{n}sec" for n in range(1, max_window + 1)]
        rolling_power_df = pd.concat(rolling_power, keys=rolling_power_names, axis=1)
        rolling_power_df.dropna(axis=1, how="all", inplace=True)
        rolling_power_df = pd.concat([self.activity[["timestamp", "power"]], rolling_power_df], axis=1)
        return rolling_power_df

    def cp_intensity(self, cp_activity=True, length: int = 1200) -> tuple[float, pd.DataFrame]:
        """Add critical power intensity to the activity dataframe
        df: dataframe with a power column
        cp_user: True=use user defined critical power, False=use calculated critical power from activity
        length: int, the length of CP curve, default is 1200 seconds
        """
        warnings.filterwarnings("ignore", category=pd.errors.PerformanceWarning)
        logging.info("Calculate critical power intensity")
        if cp_activity and not self.cp_points:
            logging.info("Calculate critical power from ride data")
            self.calculate_cp()
        logging.info("Calculate critical power intensity")
        df_rolling = self.get_power_roll_avg_df(length)
        logging.info("Calculate critical power intensity")
        for col in [col for col in df_rolling.columns if "power_" in col]:  # 1 second to 1200 seconds
            # Calculate rolling average for the interval
            interval = int(col.split("_")[1].replace("sec", ""))
            if cp_activity:
                df_rolling[f"percent_cp_{interval}sec"] = df_rolling[col] / self.cp_points[interval].cp
            else:
                df_rolling[f"percent_cp_{interval}sec"] = df_rolling[col] / self.cp_defined_dict[interval]
        df_rolling["percent_total_cp"] = (
            df_rolling[[c for c in df_rolling.columns if "percent_cp_" in c]].sum(axis=1) / length
        )
        self.activity_percent_cp = df_rolling["percent_total_cp"].mean()
        self.df_rolling = df_rolling[[col for col in df_rolling.columns if "percent_" in col]]

        return self.activity_percent_cp, self.df_rolling

    def ramp_test_activity(
        self, segment_time: int = 30, test_length: int = 1200, ftp: int = 1
    ) -> tuple[pd.DataFrame, pd.DataFrame]:
        """Convert a power profile to a ramp test workout
        The last 30sec is always 1 second ramp.
        profile: list of tuples of seconds and power [(1, 100), (2, 90), (3, 85), (4, 80)] must start with 1sec
        segment_time: int, the time in seconds of each segment of the ramp test workout, starting after 30sec
        :return: full dataframe, workout segment dataframe, workout segment dataframe with power per ftp
        """
        try:
            assert max(self.cp_defined_dict.keys()) >= test_length
        except AssertionError as err:
            raise ValueError("The profile must end with with a time >= test_length") from err

        df = self.cp_defined_df[self.cp_defined_df.seconds <= test_length].copy()

        df = _calculate_ramp_power(df)

        df["bins"] = df.apply(lambda row: row.name // segment_time + 30 if int(row.name) > 30 else row.name, axis=1)
        df["bin_power"] = df.groupby("bins")["ramp_power"].transform("mean").round(0)
        df["bin_time"] = df.groupby("bins")["seconds"].transform("count")
        df["power_per_ftp"] = df["power"] / ftp
        df["ramp_power_per_ftp"] = df["ramp_power"] / ftp
        df["bin_power_per_ftp"] = df["bin_power"] / ftp
        df["WKO Critical Power"] = df["bin_power"].expanding().mean()
        df_wko = df.drop_duplicates(subset=["bins"], keep="first")
        df_wko = df_wko[["bins", "bin_time", "bin_power", "bin_power_per_ftp"]].copy()
        df_wko.rename(
            columns={"bin_power": "power", "bin_time": "duration", "bin_power_per_ftp": "power%ftp"}, inplace=True
        )
        df_wko.sort_values(by="bins", ascending=False, inplace=True)
        df_wko.reset_index(drop=True, inplace=True)
        df_wko["segment"] = df_wko.index + 1
        self.ramp_test_df = df
        self.ramp_test_wko = df_wko[["segment", "duration", "power", "power%ftp"]]
        return df, df_wko[["segment", "duration", "power", "power%ftp"]]

    def make_zwo_from_ramp(self, workout: pd.DataFrame, filename: str | None, name: str, ftp: int | None = 1):
        xml = "<?xml version='1.0' encoding='UTF-8'?>\n"
        xml += "<workout_file>\n"
        xml += "  <author>Vincent Davis</author>\n"
        xml += f"  <name>Most Painful Ramp Test {name}</name>\n"
        xml += "  <description>A ramp test based on power profile</description>\n"
        xml += "  <sportType>bike</sportType>\n"
        xml += "  <tags></tags>\n"
        if ftp:
            xml += f"  <ftpOverride>{ftp}</ftpOverride>\n"
        xml += "  <workout>\n"
        for r in workout.to_dict(orient="records"):
            d = r["duration"]
            p = r["power%ftp"]
            xml += f'      <SteadyState Duration="{d}" Power="{p}"/>\n'
        xml += "    </workout>\n"
        xml += "</workout_file>\n"
        if filename:
            with open(filename, "w") as f:
                f.write(xml)
        return xml
